! 2D initial data solver for Brill wave / black hole spacetime
! contains hole
! Uses red/black Gauss-Seidel SOR
! Dirchelet boundary conditions
! David Rideout  Sept 1996
! to compile on Cray: f90 hole2c.F90 -o h2c -O 3

! commenting out S eliminates tPsi, which is used for convergence testing
! commenting out R prevents relaxation
! commenting out RI prevents Richardson extrapolation
! commenting out RO gives Dirchelet outer boundary conditions, but look at init_data subroutine!
define(D,!)
define(S, ) 
define(R, )
define(RI,!)
define(RO,!)

module data
implicit none
integer, parameter :: re=selected_real_kind(13,307) !nice dble precision real
integer :: nx, ny, nz, i, j, k, tmp0, tmp1, tmp2
integer, dimension(:,:), allocatable :: mask
integer, dimension(:), allocatable :: mi, mj, di, dj
integer :: nl
real(re) , dimension(:,:,:), allocatable, target :: Psi, res, f
S real(re) , dimension(:,:,:), allocatable, target :: tPsi
real(re), dimension(:,:), allocatable :: ra
real(re) :: h, xmax
real(re), dimension(:,:,:), pointer :: m
end module data
!************************************************************************
program Initial_Data
use data
implicit none
integer, parameter :: dump_iter=1 ! when residual norms are dumped
real(re), parameter :: pi=3.14159268
real(re) :: normres=10.0_re, l2n_int, l2n_ob=0, l2n_ib, tol, w_int, w_ob, a
real(re) :: error, error_int, error_ob
integer :: iter=0
character(20) :: filenam
interface
   function relax_int(w)
     use data
     implicit none
     real(re) :: relax_int, w, sumsq=0_re
     integer :: tmp
   end function relax_int
   function relax_ib
     use data
     implicit none
     real(re) :: relax_ib, sumsq, Ja
     integer :: l
   end function relax_ib
end interface

! Read parameters from input file
call parameters(w_int, w_ob, tol, a)
h=xmax/nx
print *, 'nx, ny, nz, w_int, w_ob, xmax, tol, a, h'
print *, nx, ny, nz, w_int, w_ob, xmax, tol, a, h  ! to ensure parameters read properly

! Allocate arrays
tmp0 = 2*nx !int(2*2*pi*a)+1 ! guess that nl won't exceed this
allocate(mask(-nx:nx, -ny:ny), mi(1:tmp0), mj(1:tmp0), di(1:tmp0), dj(1:tmp0))
allocate(Psi(-nx:nx,-ny:ny,-nz:nz), res(-nx:nx,-ny:ny,-nz:nz), &
  f(-nx:nx,-ny:ny,-nz:nz))
tmp0=max(nx,ny,nz)
allocate( ra(-tmp0:tmp0, -tmp0:tmp0) )
!tmp0 = 8_re*(nx*ny+nx*nz+ny*nz)+2_re  ! number ob points
!tmp1 = (2*nx-1)*(2*ny-1)*(2*nz-1)     ! number interior points
!tmp2 = (2*nx+1)*(2*ny+1)*(2*nz+1)     ! total number of points
tmp0 = 4_re*(nx + ny) ! number ob points
tmp1 = (2*nx-1)*(2*ny-1)     ! number interior points
tmp2 = (2*nx+1)*(2*ny+1)     ! total number of points

! Generate theoretical solution & source term
call init_data(a)

!m=>tPsi
!call xyplot(0,'tPsi0.p')
!call xyplot(nz-1,'tPsiz-1.p')
!call xyplot(nz,'tPsiz.p')
!stop

! Perform Gauss-Seidel iteration:
!res=0_re	! set boundary residuals to zero !should not be necessary!
!do while (abs(normres-oldnormres)> tol)
do while (normres> tol)
!   l2n_int = relax_int(w_int)
RO   l2n_ob = relax_ob(w_ob)
   l2n_ib = relax_ib()

!m=>res
!call zplot(-nx,ny,'nxpyz.p')
!call xplot(0,nz,'zp.p')
!call xplot(-ny,0,'mz.p')
!call xplot(ny,0,'pz.p')
!call xyplot(-nz+1, 'botslm1.p')
!call xyplot(-nz, 'botslice.p')
filenam = char(48+iter)//'.p'
print *, filenam
!call xzplot(0,'mid'//filenam)
!call xzplot(ny,'top'//filenam)
!if (iter==6) stop
!call xyplot(nz-1,'topm1.p')
!call xzplot(-ny,'negy.p')
!call xzplot(ny,'posy.p')
!call yzplot(-nx,'negx.p')
!call yzplot(nx,'posx.p')
!m=>f
!call yzplot(nx,'source.p')
!call yzplot(0,'midsource.p')
!stop

m=>f
!call xyplot(nx,'top'//char(48+int(iter/100))//'.p')
!call xyplot(nx-1,'topm1'//char(48+iter)//'.p')
call xyplot(0,'mid'//filenam)
call yplot(0,0,'c'//filenam)
l2n_int=5
if (iter==10) stop

!   oldnormres=normres
!   m => res
!   call l2n(normres)
!   call l2n_bdry(normresb)

!normres = sqrt((tmp0*l2n_ob**2 + tmp1*l2n_int**2)/tmp2)
normres = l2n_int  ! for now, It makes life easier


!   print *, 'overall norm of residual:', normres
!   print *, 'norm of residual on outer boundary:', normresb
!   print *, 'corners:', res(-nx,-ny,-nz), res(nx,-ny,-nz), res(-nx,ny,-nz), &
!        res(nx,ny,-nz), res(-nx,-ny,nz), res(nx,-ny,nz), res(-nx,ny,nz), &
!        res(nx,ny,nz)

   iter = iter + 1 !count number of iterations
   if (mod(iter, dump_iter)==0 .or. iter==1) then
      print *, 'iter',iter,'residue norm',normres
      print *, '   interior', l2n_int, 'ib', l2n_ib, 'ob', l2n_ob
!      m => Psi
!      filenam = char(48+int((iter-1)/100))//'.p'
!      print *, filenam
!      call xplot(ny,nz,'xpp'//filenam)
!      call xplot(-ny,-nz,'xnn'//filenam)
!      call xplot(-ny,nz,'xnp'//filenam)
!      call xplot(ny,nz,'xpn'//filenam)
!      call yplot(nx,nz,'ypp'//filenam)
!      call zplot(-nx,ny,'znp'//filenam)
   end if

   if (normres>999_re) then
      print *, 'iteration',iter,' Residuals too large.'
      stop  ! Let's not sit here forever if it is going to diverge!
   end if
end do

! Output final results:
S error=sum(abs(Psi-tPsi))/(2*nx+1)/(2*ny+1)/(2*nz+1)
                                !average deviation from theoretical solution
S error_int=sqrt(sum((Psi(-nx+1:nx-1, -ny+1:ny-1, -nz+1:nz-1) - &
     tPsi(-nx+1:nx-1, -ny+1:ny-1, -nz+1:nz-1))**2)/(2*nx-1)/(2*ny-1)/(2*nz-1))
S res = Psi - tPsi     ! to calculate error at ob
S m=> res
S call l2n_bdry(error_ob)

print *, 'x grid size:', 2*nx+1
print *, 'Solution took',iter,' iterations.'
print *, 'Norm of Residual is',normres,' .'
S print *, 'Average error from theoretical solution is', error,' .'
S print *, 'l2 norm of error in interior:', error_int
S print *, 'l2 norm of error on outer boundary:',error_ob

RI write(20) nx
RI write(30) Psi
RI write(40) tPsi
RI close(20)
RI close(30)
RI close(40)
RI print *, ''
RI print *, 'Run fine grid first; mv fort.30 to fort.50'
RI print *, 'Then run coarse grid; run rich'

! Output for convergence test
!do i=-n, n
!   if (theory) then
!      print '(F12.9, F19.15)', i*h, tPsi(i,0,0)
!   else
!      print '(F12.9, F19.15)', i*h, Psi(i,0,0)
!   endif
!end do

end program Initial_Data
!************************************************************************
subroutine parameters(w_int, w_ob, tol, a)
use data
implicit none
real(re) :: w_int, w_ob, tol, a
character(10) :: check="" ! checks for errors

open(1, file='init.in')
read(1,*) nx, ny, nz  ! (2nx+1)(2ny+1)(2nz+1) data points...
read(1,*) check
if (.not. check=="check") then
   print *,"Error in init.in ; program terminating"
   stop
end if
read(1,*) a, w_int, w_ob, xmax, tol
close(1)

end subroutine parameters
!------------------------------------------------------------------------
subroutine init_data(a)
use data
implicit none
real(re) :: a
integer :: l

! Generate theoretical solution and boundary values:
do i=-nx, nx
   do j=-ny, ny
      do k=-nz, nz
         res(i,j,k) = (i**2 + j**2 + k**2)*h**2
      end do     ! res is radius squared for now
   end do
end do
Psi=exp(-res)+1._re ! Psi holds the theoretical solution for now
S allocate(tPsi(-nx:nx,-ny:ny,-nz:nz))
S tPsi=Psi	

!Psi=1     ! Initial guess for Psi
!Psi(-nx+1:nx-1,-ny+1:ny-1,-nz+1:nz-1)=1

! Generate 'source term' in interior and boundaries:
f = 2._re*exp(-res)*(2._re*res-3._re)
do i=-nx, nx
   do j=-ny, ny
      do k=-nz, nz
         if (abs(i)==nx .or. abs(j)==ny .or. abs(k)==nz) &
              f(i,j,k)=exp(-res(i,j,k))*((1_re/sqrt(res(i,j,k))) &
              - 2_re*sqrt(res(i,j,k)))
      end do     ! res is radius squared for now
   end do
end do

! Hole:
mask = 1
do i=-nx, nx
   do j=-ny, ny
      if (res(i,j,0) < a**2) mask(i,j)=0
   end do
end do
l=1
do i=-nx+4, nx-4
   do j=-ny+4, ny-4
      if (mask(i,j)==0) then
         if (mask(i+1,j)==1 .or. mask(i-1,j)==1 .or. mask(i,j+1)==1 &
              .or. mask(i,j-1)==1) then
            mi(l)=i; mj(l)=j
            l=l+1
         end if
      end if
   end do
end do
nl=l-1
if (nl>2*nx) then
   print *, 'nl too big'; stop
end if

do l=1, nl
   if (mi(l)<0) then
      if (mj(l)<0) then
         di(l)=-1; dj(l)=-1
      else
         di(l)=-1; dj(l)=+1
      end if
   else
      if (mj(l)<0) then
         di(l)=+1; dj(l)=-1
      else
         di(l)=+1; dj(l)=+1
      end if
   end if
end do

!source term on inner boundary
do l=1, nl
   i=mi(l); j=mj(l)
   f(i,j,0) = -(4*res(i,j,0)*exp(-res(i,j,0)) + exp(-res(i,j,0)) +1)/2
end do

D do j=-ny, ny
D    print '(17I3)', mask(:, j)
D end do

D do l=1, nl
D    print *, l, mi(l), mj(l), di(l), dj(l)
D end do
D stop

end subroutine init_data
!------------------------------------------------------------------------
function relax_int(w)
use data
implicit none
real(re) :: relax_int, w, sumsq=0_re, Ja
integer :: tmp

!red
sumsq=0_re
Ja = -4/h**2
   do i=-nx+1, nx-1
      do j=-ny+1, ny-1
!         do k=-nz+1, nz-1
            k=0
            tmp = 1-abs(mod(i+j+k,2))
            res(i,j,k) = &
                ((Psi(i-1,  j,k  ) + Psi(i+1,j  ,k  ) + &
                  Psi(i  ,j-1,k  ) + Psi(i  ,j+1,k  ) + &
                  Psi(i  ,j  ,k-1) + Psi(i  ,j  ,k+1) - &
                  6*Psi(i,j,k))/h/h - f(i,j,k))
R           Psi(i,j,k) = -mask(i,j)*real(tmp,re)*w*res(i,j,k)/Ja + Psi(i,j,k)
!        end do
      end do
   end do
!black
   do i=-nx+1, nx-1
      do j=-ny+1, ny-1
!         do k=-nz+1, nz-1
            k=0
            tmp = abs(mod(i+j+k,2))
            res(i,j,k) = &
                 (Psi(i-1,  j,k  ) + Psi(i+1,j  ,k  ) + &
                  Psi(i  ,j-1,k  ) + Psi(i  ,j+1,k  ) + &
                  Psi(i  ,j  ,k-1) + Psi(i  ,j  ,k+1) - &
                  6*Psi(i,j,k))/h/h - f(i,j,k)
R           Psi(i,j,k) = real(tmp,re)*w*h*h*res(i,j,k)/6_re + Psi(i,j,k)
            sumsq = sumsq + mask(i,j)*res(i,j,k)**2
!         end do
      end do
   end do
k = tmp1 - (tmp2-sum(mask)) ! interior points less hole points
D print *, tmp1, tmp2, sum(mask), k
relax_int = sqrt(sumsq/k)

end function relax_int
!------------------------------------------------------------------------------
function relax_ib
use data
implicit none
real(re) :: relax_ib, sumsq, Ja
integer :: l

sumsq=0_re
do l=1, nl
   i=mi(l); j=mj(l)
   res(i,j,0) = (i*di(l)*(Psi(i+di(l),j,0)-Psi(i,j,0)) &
        + j*dj(l)*Psi(i,j*dj(l),0))/h - Psi(i,j,0)/2 - f(i,j,0)
   Ja = - i*di(l) - j*dj(l) - 0.5_re
   Psi(i,j,0) = -res(i,j,0)/Ja + Psi(i,j,0)
   sumsq = sumsq + res(i,j,0)**2
end do
relax_ib = sqrt(sumsq/nl)

end function relax_ib
!------------------------------------------------------------------------------
subroutine l2n(norm)
use data
implicit none
real(re) :: norm

!norm=sqrt(sum(m*m))
!norm=sqrt(sum(m**2)/((2*nx+1)*(2*ny+1)*(2*nz+1)))
norm=sum(abs(m))/((2*nx+1)*(2*ny+1)*(2*nz+1))

!siz = size(m)
!print *, 'size is', siz

end subroutine l2n
!------------------------------------------------------------------------------
subroutine l2n_bdry(norm)
use data
implicit none
real(re) :: norm

norm=0
do i=-nx, nx
   do j=-ny, ny
      do k=-nz, nz
         if (abs(i)==nx .or. abs(j)==ny .or. abs(k)==nz) &
              norm = norm + m(i,j,k)**2
      end do
   end do
end do
norm = sqrt(norm / (8_re*(nx*ny+nx*nz+ny*nz)+2_re))

end subroutine l2n_bdry
!------------------------------------------------------------------------------
subroutine dump
use data
implicit none

do k=-nz, nz
   print *, 'z index is ', k
   do j=-ny, ny
      print '(5F14.9)', m(:,j,k)
   end do
end do
print *,''

end subroutine dump
!------------------------------------------------------------------------------
subroutine xplot(y, z, name)
use data
implicit none
!real(re), dimension(:,:,:) :: fn
integer :: y, z
character*(*) name
integer :: x

open(1, file=name, status="replace", action="write")
do x=-nx, nx
   write(1, '(1P,2e14.4)') x*h, m(x,y,z)
end do
close(1)

end subroutine xplot
!------------------------------------------------------------------------------
subroutine yplot(x, z, name)
use data
implicit none
!real(re), dimension(:,:,:) :: fn
integer :: x, z
character*(*) name
integer :: y

open(1, file=name, status="replace", action="write")
do y=-ny, ny
   write(1, '(1P,2e14.4)') y*h, m(x,y,z)
end do
close(1)

end subroutine yplot
!------------------------------------------------------------------------------
subroutine zplot(x, y, name)
use data
implicit none
!real(re), dimension(:,:,:) :: fn
integer :: x, y
character*(*) name
integer :: z

open(1, file=name, status="replace", action="write")
do z=-nz, nz
   write(1, '(1P,2e14.4)') z*h, m(x,y,z)
end do
close(1)

end subroutine zplot
!------------------------------------------------------------------------------
subroutine xyplot(z, name)
use data
implicit none
!real(re), dimension(:,:,:) :: fn
integer :: z
character*(*) name
integer :: x, y

open(1, file=name, status="replace", action="write")
do x=-nx, nx
   do y=-ny, ny
      write(1, '(1P,3e14.4)') x*h, y*h, m(x,y,z)
   end do
!   write(1, *)
end do
close(1)

end subroutine xyplot
!------------------------------------------------------------------------------
subroutine xzplot(y, name)
use data
implicit none
!real(re), dimension(:,:,:) :: fn
integer :: y
character*(*) name
integer :: x, z

open(1, file=name, status="replace", action="write")
do x=-nx, nx
   do z=-nz, nz
      write(1, '(1P,3e14.4)') x*h, z*h, m(x,y,z)
   end do
!   write(1, *)
end do
close(1)

end subroutine xzplot

!------------------------------------------------------------------------------
subroutine yzplot(x, name)
use data
implicit none
!real(re), dimension(:,:,:) :: fn
integer :: x
character*(*) name
integer :: y, z

open(1, file=name, status="replace", action="write")
do y=-ny, ny
   do z=-nz, nz
      write(1, '(1P,3e14.4)') y*h, z*h, m(x,y,z)
   end do
!   write(1, *)
end do
close(1)

end subroutine yzplot
